  !! Allocate a buffer array large enough to hold data of the largest block of the domain.
  !! Allocates a pointer.
  !! 
  !! The number of bins (lower bound, upper bound) has to be specified.
  !! Additional space for ghost cells can be allocated by setting $add_ghosts.
  !!
  !! The bounds for the 3 spatial dimensions are allocated such that the block interior
  !! grid cells start at 1. E.g. if allocating with 2 ghost cells, the lower bound is -1.
  !!
  !! Uses $fd4_domain_max_bext.
  !! Returns bufp => NULL() on error.
  !!
  subroutine fd4_domain_buf_pointer(domain, bin0, bin1, add_ghosts, bufp)
    type(fd4_domain) , intent(in)  :: domain        !! the domain
    integer(i_k)     , intent(in)  :: bin0          !! lower bound for 4th dimension
    integer(i_k)     , intent(in)  :: bin1          !! upper bound for 4th dimension
    logical          , intent(in)  :: add_ghosts    !! adds ghost cells to the buffer extent?
    real(r_k)        , pointer     :: bufp(:,:,:,:) !! the pointer to allocate
    
    integer(i_k) :: bext(3), bnd(3,2), dim, ag
    
    bufp => NULL()
    FD4_L1_ASSERT(bin0>bin1,'bin0 > bin1')
    FD4_L1_ASSERT(.not. domain%init,'domain empty, create one first')
    ag = 0
    if(add_ghosts) ag = 1
    call fd4_domain_max_bext(domain, bext(1:3))
    do dim=1,3
      bnd(dim,1) =         1 - ag * domain%nghosts(dim)
      bnd(dim,2) = bext(dim) + ag * domain%nghosts(dim)
    end do
    allocate( bufp( bin0:bin1, bnd(1,1):bnd(1,2), bnd(2,1):bnd(2,2), bnd(3,1):bnd(3,2) ) )
      
  end subroutine fd4_domain_buf_pointer
  
  
  !! Allocate a buffer array large enough to hold data of the largest block of the domain.
  !! Allocates an allocatable. This is Fortran 95 TR15581, not all compilers may support it.
  !! 
  !! The number of bins (lower bound, upper bound) has to be specified.
  !! Additional space for ghost cells can be allocated by setting $add_ghosts.
  !!
  !! The bounds for the 3 spatial dimensions are allocated such that the block interior
  !! grid cells start at 1. E.g. if allocating with 2 ghost cells, the lower bound is -1.
  !!
  !! Uses $fd4_domain_max_bext.
  !! Returns unallocated buf on error.
  !!
  subroutine fd4_domain_buf_allocatable(domain, bin0, bin1, add_ghosts, buf)
    type(fd4_domain)     , intent(in)  :: domain        !! the domain
    integer(i_k)         , intent(in)  :: bin0          !! lower bound for 4th dimension
    integer(i_k)         , intent(in)  :: bin1          !! upper bound for 4th dimension
    logical              , intent(in)  :: add_ghosts    !! adds ghost cells to the buffer extent?
    real(r_k),allocatable, intent(out) :: buf(:,:,:,:)  !! the array to allocate
    
    integer(i_k) :: bext(3), bnd(3,2), dim, ag
    
    FD4_L1_ASSERT(bin0>bin1,'bin0 > bin1')
    FD4_L1_ASSERT(.not. domain%init,'domain empty, create one first')
    ag = 0
    if(add_ghosts) ag = 1
    call fd4_domain_max_bext(domain, bext(1:3))
    do dim=1,3
      bnd(dim,1) =         1 - ag * domain%nghosts(dim)
      bnd(dim,2) = bext(dim) + ag * domain%nghosts(dim)
    end do
    allocate( buf( bin0:bin1, bnd(1,1):bnd(1,2), bnd(2,1):bnd(2,2), bnd(3,1):bnd(3,2) ) )      

  end subroutine fd4_domain_buf_allocatable
